__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from jina.parsers.peapods.runtimes.distributed import mixin_distributed_feature_parser


def set_pea_parser(parser=None):
    if not parser:
        from .base import set_base_parser
        parser = set_base_parser()

    from .peapods.base import mixin_base_ppr_parser
    from .peapods.runtimes.zmq import mixin_zmq_runtime_parser
    from .peapods.runtimes.zed import mixin_zed_runtime_parser
    from .peapods.runtimes.container import mixin_container_runtime_parser
    from .peapods.runtimes.remote import mixin_remote_parser
    from .peapods.pea import mixin_pea_parser

    mixin_base_ppr_parser(parser)
    mixin_zmq_runtime_parser(parser)
    mixin_zed_runtime_parser(parser)
    mixin_container_runtime_parser(parser)
    mixin_remote_parser(parser)
    mixin_distributed_feature_parser(parser)
    mixin_pea_parser(parser)

    return parser


def set_pod_parser(parser=None):
    if not parser:
        from .base import set_base_parser
        parser = set_base_parser()

    set_pea_parser(parser)

    from .peapods.pod import mixin_base_pod_parser

    mixin_base_pod_parser(parser)

    return parser


def set_gateway_parser(parser=None):
    if not parser:
        from .base import set_base_parser
        parser = set_base_parser()

    from .peapods.base import mixin_base_ppr_parser
    from .peapods.runtimes.zmq import mixin_zmq_runtime_parser
    from .peapods.runtimes.zed import mixin_zed_runtime_parser
    from .peapods.runtimes.container import mixin_container_runtime_parser
    from .peapods.runtimes.remote import mixin_remote_parser
    from .peapods.runtimes.remote import mixin_grpc_parser
    from .peapods.pea import mixin_pea_parser

    mixin_base_ppr_parser(parser)
    mixin_zmq_runtime_parser(parser)
    mixin_zed_runtime_parser(parser)
    mixin_grpc_parser(parser)
    mixin_remote_parser(parser)
    mixin_pea_parser(parser)

    from ..enums import SocketType, PodRoleType

    parser.set_defaults(name='gateway',
                        socket_in=SocketType.PULL_CONNECT,  # otherwise there can be only one client at a time
                        socket_out=SocketType.PUSH_CONNECT,
                        ctrl_with_ipc=True,  # otherwise ctrl port would be conflicted
                        read_only=True,
                        runtime_cls='GRPCRuntime',
                        pod_role=PodRoleType.GATEWAY)
    return parser


def set_client_cli_parser(parser=None):
    if not parser:
        from .base import set_base_parser
        parser = set_base_parser()

    from .peapods.runtimes.remote import mixin_grpc_parser, mixin_remote_parser
    from .client import mixin_client_cli_parser

    mixin_client_cli_parser(parser)
    mixin_grpc_parser(parser)
    mixin_remote_parser(parser)

    return parser


def get_main_parser():
    from .base import set_base_parser
    from .helloworld import set_hw_parser, set_hw_chatbot_parser
    from .helper import _chf, _SHOW_ALL_ARGS
    from .check import set_check_parser
    from .export_api import set_export_api_parser
    from .flow import set_flow_parser
    from .hub import set_hub_parser
    from .logger import set_logger_parser
    from .ping import set_ping_parser
    from .optimizer import set_optimizer_parser

    # create the top-level parser
    parser = set_base_parser()

    sp = parser.add_subparsers(dest='cli',
                               description='use `%(prog)-8s [sub-command] --help` '
                                           'to get detailed information about each sub-command', required=True)

    set_hw_parser(sp.add_parser('hello-world',
                                help='👋 Hello World! Hello Jina!',
                                description='Start the hello-world demo, a simple end2end image index and search demo '
                                            'without any extra dependencies.',
                                formatter_class=_chf))

    set_pod_parser(sp.add_parser('pod',
                                 help='Start a Pod',
                                 description='Start a Jina Pod',
                                 formatter_class=_chf))

    set_flow_parser(sp.add_parser('flow',
                                  description='Start a Flow that orchestrates multiple pods',
                                  help='Start a Flow',
                                  formatter_class=_chf))

    set_optimizer_parser(sp.add_parser('optimizer',
                                       description='Start a FlowOptimizer from a YAML configuration file',
                                       help='Start an FlowOptimizer from a YAML file', formatter_class=_chf))

    set_gateway_parser(sp.add_parser('gateway',
                                     description='Start a Gateway that receives client Requests via gRPC/REST interface',
                                     help='Start a Gateway',
                                     formatter_class=_chf))

    set_ping_parser(sp.add_parser('ping',
                                  help='Ping a pod and check its connectivity',
                                  description='Ping a remote pod and check the network connectivity',
                                  formatter_class=_chf))

    set_check_parser(sp.add_parser('check',
                                   help='Check the import of all Executors and Drivers',
                                   description='Check the import status of all executors and drivers',
                                   formatter_class=_chf))

    set_hub_parser(sp.add_parser('hub', help='Build, push, pull Jina Hub images',
                                 description='Build, push, pull Jina Hub images',
                                 formatter_class=_chf))

    # Below are low-level / internal / experimental CLIs, hidden from users by default

    set_pea_parser(sp.add_parser('pea',
                                 description='Start a Jina pea. '
                                             'You should rarely use this directly unless you '
                                             'are doing low-level orchestration',
                                 formatter_class=_chf, **(dict(help='start a pea')) if _SHOW_ALL_ARGS else {}))

    set_logger_parser(sp.add_parser('log',
                                    description='Receive piped log output and beautify the log. '
                                                'Depreciated, use Jina Dashboard instead',
                                    formatter_class=_chf,
                                    **(dict(help='beautify the log')) if _SHOW_ALL_ARGS else {}))

    set_client_cli_parser(sp.add_parser('client',
                                        description='Start a Python client that connects to a remote Jina gateway',
                                        formatter_class=_chf,
                                        **(dict(help='start a client')) if _SHOW_ALL_ARGS else {}))

    set_export_api_parser(sp.add_parser('export-api',
                                        description='Export Jina API to JSON/YAML file for 3rd party applications',
                                        formatter_class=_chf,
                                        **(dict(help='export Jina API to file')) if _SHOW_ALL_ARGS else {}))

    set_hw_chatbot_parser(sp.add_parser('hello-world-chatbot',
                                        **(dict(help='Covid-19 chatbot based on DistilBERT')) if _SHOW_ALL_ARGS else {},
                                        description='Start a hello-world demo: a simple Covid-19 chatbot. '
                                                    'Pytorch and transformers are required to run this demo',
                                        formatter_class=_chf))
    return parser



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import warnings
from typing import Optional

from . import BaseExecutableDriver, FastRecursiveMixin
from ..types.sets import DocumentSet


class BaseEncodeDriver(BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`encode` by default """

    def __init__(self,
                 executor: str = None,
                 method: str = 'encode',
                 *args, **kwargs):
        super().__init__(executor, method, *args, **kwargs)


class EncodeDriver(FastRecursiveMixin, BaseEncodeDriver):
    """Extract the content from documents and call executor and do encoding
    """

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        contents, docs_pts = docs.all_contents

        if docs_pts:
            embeds = self.exec_fn(contents)
            if len(docs_pts) != embeds.shape[0]:
                self.logger.error(
                    f'mismatched {len(docs_pts)} docs from level {docs_pts[0].granularity} '
                    f'and a {embeds.shape} shape embedding, the first dimension must be the same')
            for doc, embedding in zip(docs_pts, embeds):
                doc.embedding = embedding


class LegacyEncodeDriver(BaseEncodeDriver):
    """Extract the content from documents and call executor and do encoding
    """

    class CacheDocumentSet:
        """Helper class to accumulate documents from differents document Set in a single DocumentSet
         to help guarantee that the encoder driver can consume documents in fixed batch sizes to allow
         the EncoderExecutors to leverage its batching abilities.
         It is useful to have batching even when chunks are involved"""

        def __init__(self,
                     capacity: Optional[int] = None,
                     *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.capacity = capacity
            self._doc_set = DocumentSet(docs_proto=[])

        @property
        def available_capacity(self):
            return self.capacity - len(self._doc_set)

        def cache(self, docs: DocumentSet):
            docs_to_append = min(len(docs), self.available_capacity)
            self._doc_set.extend(docs[: docs_to_append])
            return DocumentSet(docs[docs_to_append:])

        def __len__(self):
            return len(self._doc_set)

        def get(self):
            return self._doc_set

    def __init__(self,
                 batch_size: Optional[int] = None,
                 *args, **kwargs):
        """
        Extract from different documents.
        :param batch_size: number of documents to be used simultaneously in the encoder :meth:_apply_all.
        It is specially useful when the same EncoderExecutor can be used for documents of different granularities
         (chunks, chunks of chunks ...)
        .. warning::
            - This parameter was added to cover the case where root documents had very few chunks, and the encoder executor could
            then only process them in batches of the chunk size of each document, which did not lead to the full use of batching capabilities
            of the powerful Executors
        """
        super().__init__(*args, **kwargs)
        warnings.warn(f'this drivers will be removed soon, use {EncodeDriver!r} instead', DeprecationWarning)
        self.batch_size = batch_size
        if self.batch_size:
            self.cache_set = LegacyEncodeDriver.CacheDocumentSet(capacity=self.batch_size)
        else:
            self.cache_set = None

    def __call__(self, *args, **kwargs):
        self._traverse_apply(self.docs, *args, **kwargs)
        self._empty_cache()

    def _apply_batch(self, batch: 'DocumentSet'):
        contents, docs_pts = batch.all_contents

        if docs_pts:
            embeds = self.exec_fn(contents)
            if embeds is None:
                self.logger.error(
                    f'{self.exec_fn!r} returns nothing, you may want to check the implementation of {self.exec!r}')
            elif len(docs_pts) != embeds.shape[0]:
                self.logger.error(
                    f'mismatched {len(docs_pts)} docs from level {docs_pts[0].granularity} '
                    f'and a {embeds.shape} shape embedding, the first dimension must be the same')
            else:
                for doc, embedding in zip(docs_pts, embeds):
                    doc.embedding = embedding

    def _empty_cache(self):
        if self.batch_size:
            cached_docs = self.cache_set.get()
            if len(cached_docs) > 0:
                self._apply_batch(cached_docs)
            self.cache_set = LegacyEncodeDriver.CacheDocumentSet(capacity=self.batch_size)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        if self.cache_set is not None:
            left_docs = self.cache_set.cache(docs)
            while len(left_docs) > 0:
                self._empty_cache()
                left_docs = self.cache_set.cache(left_docs)
            if self.cache_set.available_capacity == 0:
                self._empty_cache()
        else:
            self._apply_batch(docs)



from .base import set_base_parser


def set_ping_parser(parser=None):
    if not parser:
        parser = set_base_parser()

    parser.add_argument('host', type=str,
                        help='The host address of the target Pea, e.g. 0.0.0.0')
    parser.add_argument('port', type=int,
                        help='The control port of the target pod/pea')
    parser.add_argument('--timeout', type=int, default=3000,
                        help='''
Timeout in millisecond of one check
-1 for waiting forever
''')
    parser.add_argument('--retries', type=int, default=3,
                        help='The max number of tried health checks before exit with exit code 1')
    parser.add_argument('--print-response', action='store_true', default=False,
                        help='If set, print the response when received')
    return parser



import os
import inspect
from typing import Dict, Any, Type, Set
from functools import reduce

from ..base import VersionedYAMLParser
from ....executors import BaseExecutor, get_default_metas
from ....executors.compound import CompoundExecutor


class LegacyParser(VersionedYAMLParser):
    version = 'legacy'  # the version number this parser designed for

    @staticmethod
    def _get_all_arguments(class_):
        """

        :param class_: target class from which we want to retrieve arguments
        :return: all the arguments of all the classes from which `class_` inherits
        """
        def get_class_arguments(class_):
            """
            :param class_: the class to check
            :return: a list containing the arguments from `class_`
            """
            signature = inspect.signature(class_.__init__)
            class_arguments = [p.name for p in signature.parameters.values()]
            return class_arguments

        def accumulate_classes(cls) -> Set[Type]:
            """
            :param cls: the class to check
            :return: all classes from which cls inherits from
            """
            def _accumulate_classes(c, cs):
                cs.append(c)
                if cls == object:
                    return cs
                for base in c.__bases__:
                    _accumulate_classes(base, cs)
                return cs
            
            classes = []
            _accumulate_classes(cls, classes)
            return set(classes)

        all_classes = accumulate_classes(class_)
        args = list(map(lambda x: get_class_arguments(x), all_classes))
        return set(reduce(lambda x,y: x+y,args))

    @staticmethod
    def _get_dump_path_from_config(meta_config: Dict):
        if 'name' in meta_config:
            work_dir = meta_config['workspace']
            name = meta_config['name']
            pea_id = meta_config['pea_id']
            if work_dir:
                # then try to see if it can be loaded from its regular expected workspace (ref_indexer)
                dump_path = BaseExecutor.get_shard_workspace(work_dir, name, pea_id)
                bin_dump_path = os.path.join(dump_path, f'{name}.bin')
                if os.path.exists(bin_dump_path):
                    return bin_dump_path

            root_work_dir = meta_config['root_workspace']
            root_name = meta_config['root_name']
            if root_name != name:
                # try to load from the corresponding file as if it was a CompoundExecutor, if the `.bin` does not exist,
                # we should try to see if from its workspace can be loaded as it may be a `ref_indexer`
                compound_work_dir = CompoundExecutor.get_component_workspace_from_compound_workspace(root_work_dir,
                                                                                                     root_name,
                                                                                                     pea_id)
                dump_path = BaseExecutor.get_shard_workspace(compound_work_dir, name, pea_id)
                bin_dump_path = os.path.join(dump_path, f'{name}.{"bin"}')
                if os.path.exists(bin_dump_path):
                    return bin_dump_path

    def parse(self, cls: Type['BaseExecutor'], data: Dict) -> 'BaseExecutor':
        """
        :param cls: target class type to parse into, must be a :class:`JAMLCompatible` type
        :param data: flow yaml file loaded as python dict
        :return: the Flow YAML parser given the syntax version number
        """
        _meta_config = get_default_metas()
        _meta_config.update(data.get('metas', {}))
        if _meta_config:
            data['metas'] = _meta_config

        dump_path = self._get_dump_path_from_config(data.get('metas', {}))
        load_from_dump = False
        if dump_path:
            obj = cls.load(dump_path)
            obj.logger.success(f'restore {cls.__name__} from {dump_path}')
            # consider the case where `dump_path` is not based on `obj.workspace`. This is needed
            # for
            workspace_loaded_from = data.get('metas', {})['workspace']
            workspace_in_dump = getattr(obj, 'workspace', None)
            if workspace_in_dump != workspace_loaded_from:
                obj.workspace = workspace_loaded_from
            load_from_dump = True
        else:
            cls._init_from_yaml = True

            if cls.store_args_kwargs:
                p = data.get('with', {})  # type: Dict[str, Any]
                a = p.pop('args') if 'args' in p else ()
                k = p.pop('kwargs') if 'kwargs' in p else {}
                # maybe there are some hanging kwargs in "parameters"
                # tmp_a = (expand_env_var(v) for v in a)
                # tmp_p = {kk: expand_env_var(vv) for kk, vv in {**k, **p}.items()}
                tmp_a = a
                tmp_p = {kk: vv for kk, vv in {**k, **p}.items()}
                obj = cls(*tmp_a, **tmp_p, metas=data.get('metas', {}), requests=data.get('requests', {}))
            else:
                # tmp_p = {kk: expand_env_var(vv) for kk, vv in data.get('with', {}).items()}
                obj = cls(**data.get('with', {}), metas=data.get('metas', {}), requests=data.get('requests', {}))
            cls._init_from_yaml = False

            # check if the yaml file used to instanciate 'cls' has arguments that are not in 'cls'
            arguments_from_cls = LegacyParser._get_all_arguments(cls)
            arguments_from_yaml = set(data.get('with', {}))
            difference_set = arguments_from_yaml  - arguments_from_cls
            if any(difference_set):
                obj.logger.warning(f'The arguments {difference_set} defined in the YAML are not expected in the '
                                   f'class {cls.__name__}')

            obj.logger.success(f'successfully built {cls.__name__} from a yaml config')

        # if node.tag in {'!CompoundExecutor'}:
        #     os.environ['JINA_WARN_UNNAMED'] = 'YES'

        if not _meta_config:
            obj.logger.warning(
                '"metas" config is not found in this yaml file, '
                'this map is important as it provides an unique identifier when '
                'persisting the executor on disk.')

        # for compound executor
        if not load_from_dump and 'components' in data:
            obj.components = lambda: data['components']

        obj.is_updated = False
        return obj

    def dump(self, data: 'BaseExecutor') -> Dict:
        """
        :param data: versioned executor object
        :return: the dictionary given a versioned flow object
        """
        # note: we only save non-default property for the sake of clarity
        _defaults = get_default_metas()
        p = {k: getattr(data, k) for k, v in _defaults.items() if getattr(data, k) != v}
        a = {k: v for k, v in data._init_kwargs_dict.items() if k not in _defaults}
        r = {}
        if a:
            r['with'] = a
        if p:
            r['metas'] = p

        if hasattr(data, '_drivers'):
            r['requests'] = {'on': data._drivers}

        if hasattr(data, 'components'):
            r['components'] = data.components
        return r



from ..base import set_base_parser
from ..helper import _chf


def set_hub_pushpull_parser(parser=None):
    if not parser:
        parser = set_base_parser()

    from .login import mixin_hub_docker_login_parser
    from .pushpull import mixin_hub_pushpull_parser

    mixin_hub_docker_login_parser(parser)
    mixin_hub_pushpull_parser(parser)
    return parser


def set_hub_build_parser(parser=None):
    if not parser:
        parser = set_base_parser()

    from .login import mixin_hub_docker_login_parser
    from .build import mixin_hub_build_parser

    mixin_hub_docker_login_parser(parser)
    mixin_hub_build_parser(parser)

    return parser


def set_hub_list_parser(parser=None):
    if not parser:
        parser = set_base_parser()

    from .list import mixin_hub_list_parser
    mixin_hub_list_parser(parser)

    return parser


def set_hub_new_parser(parser=None):
    if not parser:
        parser = set_base_parser()

    from .new import mixin_hub_new_parser
    mixin_hub_new_parser(parser)

    return parser


def set_hub_parser(parser=None):
    if not parser:
        parser = set_base_parser()

    spp = parser.add_subparsers(dest='hub',
                                description='use `%(prog)-8s [sub-command] --help` '
                                            'to get detailed information about each sub-command', required=True)

    spp.add_parser('login', help='login via Github to push images to Jina hub registry',
                   description='Login via Github to push images to Jina hub registry',
                   formatter_class=_chf)

    set_hub_new_parser(
        spp.add_parser('new', aliases=['init', 'create'], help='create a new Hub executor or app using cookiecutter',
                       description='Create a new Hub executor or app using cookiecutter',
                       formatter_class=_chf))

    set_hub_build_parser(
        spp.add_parser('build', help='build a directory into Jina hub image',
                       description='Build a directory into Jina hub image',
                       formatter_class=_chf))

    set_hub_pushpull_parser(
        spp.add_parser('push', help='push an image to the Jina hub registry',
                       description='Push an image to the Jina hub registry',
                       formatter_class=_chf))

    set_hub_pushpull_parser(
        spp.add_parser('pull', help='pull an image from the Jina hub registry to local',
                       description='Pull an image to the Jina hub registry to local',
                       formatter_class=_chf))

    set_hub_list_parser(
        spp.add_parser('list', help='list hub executors from jina hub registry',
                       description='List hub executors from jina hub registry',
                       formatter_class=_chf))



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple, Optional

from . import BaseExecutableDriver
from ..types.document import Document

if False:
    from .. import DocumentSet


class SegmentDriver(BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`segment` by default """

    def __init__(
            self,
            executor: Optional[str] = None,
            method: str = 'segment',
            traversal_paths: Tuple[str] = ('r',),
            *args,
            **kwargs
    ):
        super().__init__(executor, method, traversal_paths=traversal_paths, *args, **kwargs)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs):
        for doc in docs:
            _args_dict = doc.get_attrs(*self.exec.required_keys)
            ret = self.exec_fn(**_args_dict)
            if ret:
                self._update(doc, ret)

    @staticmethod
    def _update(doc, ret):
        for r in ret:
            with Document(length=len(ret), **r) as c:
                if not c.mime_type:
                    c.mime_type = doc.mime_type
            doc.chunks.append(c)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple

from . import BaseExecutableDriver, QuerySetReader, FastRecursiveMixin
from ..types.document import Document
from ..types.score import NamedScore

if False:
    from ..types.sets import DocumentSet


class BaseSearchDriver(BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`craft` by default """

    def __init__(
            self,
            executor: str = None,
            method: str = 'query',
            traversal_paths: Tuple[str] = ('r', 'c'),
            *args,
            **kwargs):
        super().__init__(
            executor,
            method,
            traversal_paths=traversal_paths,
            *args,
            **kwargs
        )


class KVSearchDriver(BaseSearchDriver):
    """Fill in the results using the :class:`jina.executors.indexers.meta.BinaryPbIndexer`

    .. warning::
        This driver runs a query for each document.
        This may not be very efficient, as the total number of queries grows cubic with the number of documents, chunks
        per document and top-k.

            - traversal_paths = ['m'] => D x K
            - traversal_paths = ['r'] => D
            - traversal_paths = ['cm'] => D x C x K
            - traversal_paths = ['m', 'cm'] => D x K + D x C x K

        where:
            - D is the number of queries
            - C is the number of chunks per document
            - K is the top-k
    """

    def __init__(self, is_update: bool = True, traversal_paths: Tuple[str] = ('m'), *args, **kwargs):
        """Construct the driver.

        :param is_update: when set to true the retrieved docs are merged into current message;
            otherwise, the retrieved Document overrides the existing Document
        :param traversal_paths: traversal paths for the driver
        :param *args: *args for super
        :param **kwargs: **kwargs for super
        """
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)
        self._is_update = is_update

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        miss_idx = []  #: missed hit results, some search may not end with results. especially in shards
        for idx, retrieved_doc in enumerate(docs):
            serialized_doc = self.exec_fn(retrieved_doc.id)
            if serialized_doc:
                r = Document(serialized_doc)
                if self._is_update:
                    retrieved_doc.update(r)
                else:
                    retrieved_doc.CopyFrom(r)
            else:
                miss_idx.append(idx)
        # delete non-existed matches in reverse
        for j in reversed(miss_idx):
            del docs[j]


class VectorFillDriver(QuerySetReader, BaseSearchDriver):
    """Fill in the embedding by their document id.
    """

    def __init__(self, executor: str = None, method: str = 'query_by_key', *args, **kwargs):
        super().__init__(executor, method, *args, **kwargs)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        embeds = self.exec_fn([d.id for d in docs])
        for doc, embedding in zip(docs, embeds):
            doc.embedding = embedding


class VectorSearchDriver(FastRecursiveMixin, QuerySetReader, BaseSearchDriver):
    """Extract embeddings from the request for the executor to query.
    """

    def __init__(self, top_k: int = 50, fill_embedding: bool = False, *args, **kwargs):
        """Construct the driver.

        :param top_k: top-k document ids to retrieve
        :param fill_embedding: fill in the embedding of the corresponding doc,
                this requires the executor to implement :meth:`query_by_key`
        :param *args: *args for super
        :param **kwargs: **kwargs for super
        """
        super().__init__(*args, **kwargs)
        self._top_k = top_k
        self._fill_embedding = fill_embedding

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        embed_vecs, doc_pts = docs.all_embeddings

        if not doc_pts:
            return

        fill_fn = getattr(self.exec, 'query_by_key', None)
        if self._fill_embedding and not fill_fn:
            self.logger.warning(f'"fill_embedding=True" but {self.exec} does not have "query_by_key" method')

        idx, dist = self.exec_fn(embed_vecs, top_k=int(self.top_k))

        op_name = self.exec.__class__.__name__
        for doc, topks, scores in zip(doc_pts, idx, dist):

            topk_embed = fill_fn(topks) if (self._fill_embedding and fill_fn) else [None] * len(topks)
            for numpy_match_id, score, vec in zip(topks, scores, topk_embed):
                m = Document(id=numpy_match_id)
                m.score = NamedScore(op_name=op_name,
                                     value=score)
                r = doc.matches.append(m)
                if vec is not None:
                    r.embedding = vec



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
import sys
import pickle

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTorchEncoder
from jina.excepts import PretrainedModelFileDoesNotExist


class TirgImageEncoder(BaseTorchEncoder):

    def __init__(self, model_path: str = 'checkpoint.pth',
                 texts_path: str = 'texts.pkl',
                 channel_axis: int = -1, 
                 *args, **kwargs):
        """
        :param model_path: the path where the model is stored.
        """
        super().__init__(*args, **kwargs)
        self.model_path = model_path
        self.texts_path = texts_path
        self.channel_axis = channel_axis
        # axis 0 is the batch
        self._default_channel_axis = 1

    def post_init(self):
        super().post_init()
        from .img_text_composition_models import TIRG
        import torch
        if self.model_path and os.path.exists(self.model_path):
            with open (self.texts_path, 'rb') as fp:
                texts = pickle.load(fp)
            self.model = TIRG(texts, 512)
            model_sd = torch.load(self.model_path, map_location=torch.device('cpu'))
            self.model.load_state_dict(model_sd['model_state_dict'])
            self.model.eval()
            self.to_device(self.model)
        else:
            raise PretrainedModelFileDoesNotExist(f'model {self.model_path} does not exist')

    def _get_features(self, data):
        return self.model.extract_img_feature(data)

    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        import numpy as np
        if self.channel_axis != self._default_channel_axis:
            data = np.moveaxis(data, self.channel_axis, self._default_channel_axis)
        import torch
        _input = torch.from_numpy(data.astype('float32'))
        if self.on_gpu:
            _input = _input.cuda()
        _feature = self._get_features(_input).detach()
        if self.on_gpu:
            _feature = _feature.cpu()
        _feature = _feature.numpy()
        return _feature